/*
 * Contributors :
 *   Pierre PETERLONGO, pierre.peterlongo@inria.fr [12/06/13]
 *   Nicolas MAILLET, nicolas.maillet@inria.fr     [12/06/13]
 *   Guillaume Collet, guillaume@gcollet.fr        [22/01/14]
 *
 * This software is a computer program whose purpose is to find all the
 * similar reads between two set of NGS reads. It also provide a similarity
 * score between the two samples.
 *
 * Copyright (C) 2014  INRIA
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "file_manager.h"
#include "file_manager_max.h"
#include "bloom_filter.h"
#include "boolean_vector.h"

#include <iostream>
#include <string>
#include <ctime>
#include <map>

std::string version = "2.0.2";

// -----------------------------------------------------------------------
//                              PROTOTYPE
// -----------------------------------------------------------------------
void print_usage ();
unsigned long compare (FileManager * index_file_manager, FileManager * search_file_manager, int kmer_size, int min_hits, unsigned long max_kmer);

// -----------------------------------------------------------------------
//                                MAIN
// -----------------------------------------------------------------------

int main (int argc, char ** argv)
{
	const clock_t begin_time = clock();
	
	FileManager * A_file_manager;
	FileManager * B_file_manager;
		
	std::map<std::string, std::string> A_files;
	std::map<std::string, std::string> B_files;
	
	int kmer_size = 32;
	int min_hits = 2;
	unsigned long max_nb_reads = 0;
	unsigned long max_kmer = 1000000000 / pow (2, 33 - kmer_size);
	
	std::string log_path;
	std::string out_path;
	
	////////////////////////////////////////////////////////////
	// Read command line arguments
	//
	std::string tmp;
	int arg_pos = 1;
	while (arg_pos < argc){
		std::string flag = argv[arg_pos];
		if (flag.compare("-a") == 0) {
			// File to index name, optionaly followed by a boolean vector file after a comma
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			tmp = argv[arg_pos];
			if (tmp.find(',') < tmp.size()) {
				A_files[tmp.substr(0, tmp.find(','))] = tmp.substr(tmp.find(',') + 1);
				std::cerr << "file A: " << tmp.substr(0, tmp.find(',')) << "\n";
			} else {
				A_files[tmp] = "";
				std::cerr << "file A: " << tmp << "\n";
			}
		} else if (flag.compare("-b") == 0) {
			// File to search name, optionaly followed by a boolean vector file after a comma
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			tmp = argv[arg_pos];
			if (tmp.find(',') < tmp.size()) {
				B_files[tmp.substr(0, tmp.find(','))] = tmp.substr(tmp.find(',') + 1);
				std::cerr << "file B: " << tmp.substr(0, tmp.find(',')) << "\n";
			} else {
				B_files[tmp] = "";
				std::cerr << "file B: " << tmp << "\n";
			}
		} else if (flag.compare("-l") == 0) {
			// The path to the log directory
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			log_path = argv[arg_pos];
		} else if (flag.compare("-o") == 0) {
			// The path to the ouput directory
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			out_path = argv[arg_pos];
		} else if (flag.compare("-k") == 0) {
			// The size of k-mers
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			kmer_size = atoi(argv[arg_pos]);
			max_kmer = 1000000000 / pow (2, 33 - kmer_size);
			std::cerr << "k-mer size (-k) = " << kmer_size << "\n";
		} else if (flag.compare("-t") == 0) {
			// The minimal number of hits to consider two reads as similar
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			min_hits = atoi(argv[arg_pos]);
			std::cerr << "min hits (-t) = " << min_hits << "\n";
		} else if (flag.compare("-m") == 0) {
			// The maximum number of reads per file (m first reads)
			arg_pos++;
			if (arg_pos >= argc) {
				std::cerr << "Error, flag " << argv[arg_pos - 1] << " needs an argument\n";
				print_usage();
				exit(1);
			}
			max_nb_reads = atoi(argv[arg_pos]);
			std::cerr << "max reads (-m) = " << max_nb_reads << "\n";
		} else if (flag.compare("-h") == 0) {
			print_usage ();
			return 0;
		} else if (flag.compare("-v") == 0) {
			std::cerr << "\ncompare_reads version " << version << "\n";
			return 0;
		} else {
			std::cerr << "Unknown option " << flag << "\n";
            print_usage ();
			return (0);
		}
		arg_pos++;
	}
	
	
	////////////////////////////////////////////////////////////
	// Check the existence of at least one file to index and
	// one file to search
	//
	if (A_files.empty())
	{
		fprintf(stderr,"No file specified with -a\n");
		print_usage ();
		return 1;
	}
	
	if (B_files.empty())
	{
		fprintf(stderr,"No file specified with -b\n");
		print_usage ();
		return 1;
	}
	
	if (max_nb_reads > 0) {
		A_file_manager = new FileManagerMax (max_nb_reads);
		B_file_manager = new FileManagerMax (max_nb_reads);
	} else {
		A_file_manager = new FileManager ();
		B_file_manager = new FileManager ();
	}
	
	for (std::map<std::string, std::string>::iterator it = A_files.begin(); it != A_files.end(); it++) {
		if (it->second.empty()) {
			A_file_manager->addFile(it->first);
		} else {
			A_file_manager->addFile(it->first, it->second);
		}
	}
	for (std::map<std::string, std::string>::iterator it = B_files.begin(); it != B_files.end(); it++) {
		if (it->second.empty()) {
			B_file_manager->addFile(it->first);
		} else {
			B_file_manager->addFile(it->first, it->second);
		}
	}
	
	
	////////////////////////////////////////////////////////////
	// Generate the name of log file from log_path (-l)
	// Open log_file and write log header
	//
	std::ofstream log_file;
	std::string log_file_name;
	if (log_path.empty()) {
		log_path = ".";
	}
	log_file_name = log_path + "/" + A_file_manager->get_sum_of_file_names() + "_VS_" + B_file_manager->get_sum_of_file_names() + ".log";
	log_file.open(log_file_name.c_str());
	
	if (!log_file.good()) {
		std::cerr << "Cannot open file " << log_file_name << " for writing\n";
	}
	
	////////////////////////////////////////////////////////////
	// If out_path is empty then the directory
	// of the first index file is used
	//
	if (out_path.empty()) {
		if (A_files.begin()->first.rfind("/") < A_files.begin()->first.size()) {
			out_path = A_files.begin()->first.substr(0, A_files.begin()->first.rfind("/"));
		} else {
			out_path = ".";
		}
	}
	
	log_file << "Compareads " << version << "\n";
	log_file << "Two reads are similar if they share at least " << min_hits << " k-mers of " << kmer_size << " nucleotides\n";
	log_file.precision(5);
	log_file << "Parameters: t=" << min_hits << " -- k=" << kmer_size << " (using " << (float) pow(2, kmer_size - 1)/(float) 1073741824 << " Gb), Max kmers in BDS=" << max_kmer << " \n";
	log_file << "Total number of reads in " << B_file_manager->get_sum_of_file_names() << ": " << B_file_manager->get_total_nb_reads() << "\n";
	
	const std::vector<std::string> & A_file_names = A_file_manager->get_file_names();
	const std::vector<std::string> & B_file_names = B_file_manager->get_file_names();
	
	unsigned long nb_total_reads = A_file_manager->get_total_nb_reads() + B_file_manager->get_total_nb_reads();
	
	unsigned long nb_common = 0;
	////////////////////////////////////////////////////////////
	// First step : B in A
	//
	
	// output message
	std::cerr << "------------------------------------------------------------------\n";
	std::cerr << "finding reads from {";
	int i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}\n";
	std::cerr << "------------------------------------------------------------------\n";
	
	// Compare first step
	compare (A_file_manager, B_file_manager, kmer_size, min_hits, max_kmer);
	
	////////////////////////////////////////////////////////////
	// Second step A in [B in A]
	//
	B_file_manager->apply_bv_on_files();
	A_file_manager->rewind();
	B_file_manager->rewind();
	
	// output message
	std::cerr << "------------------------------------------------------------------\n";
	std::cerr << "finding reads from {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "} present in [reads from {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}]\n";
	std::cerr << "------------------------------------------------------------------\n";
	
	// Compare second step
	nb_common += compare (B_file_manager, A_file_manager, kmer_size, min_hits, max_kmer);
	A_file_manager->save_bv(out_path, B_file_manager->get_sum_of_file_names());
	
	////////////////////////////////////////////////////////////
	// third step [B in A] in [A in [B in A]]
	//
	A_file_manager->apply_bv_on_files();
	A_file_manager->rewind();
	B_file_manager->rewind();
	
	// ouput message
	std::cerr << "------------------------------------------------------------------\n";
	std::cerr << "finding [reads from {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}] present in [reads from {";
		i = 0;
		for (; i < (int) A_file_names.size() - 1; i++) {
			std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
		}
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "} present in [reads from {";
		i = 0;
		for (; i < (int) B_file_names.size() - 1; i++) {
			std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
		}
		std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} present in raw {";
		i = 0;
		for (; i < (int) A_file_names.size() - 1; i++) {
			std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
		}
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}]]\n";
	std::cerr << "------------------------------------------------------------------\n";
	
	// Compare third step
	nb_common += compare (A_file_manager, B_file_manager, kmer_size, min_hits, max_kmer);
	B_file_manager->save_bv(out_path, A_file_manager->get_sum_of_file_names());
	log_file.close();
	
	std::cerr << "\n-------------------------------------------\n";
	std::cerr << "COMPUTATION DONE FOR {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "} VS {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "}\n";
	std::cerr << "Results are in\n";
	for (std::map<std::string, std::string>::iterator it = A_files.begin(); it != A_files.end(); it++) {
		std::cerr << "\t" << out_path << "/" << it->first.substr(it->first.rfind('/') + 1) << "_in_" << B_file_manager->get_sum_of_file_names() << ".bv\n";
	}
	std::cerr << "and\n";
	for (std::map<std::string, std::string>::iterator it = B_files.begin(); it != B_files.end(); it++) {
		std::cerr << "\t" << out_path << "/" << it->first.substr(it->first.rfind('/') + 1) << "_in_" << A_file_manager->get_sum_of_file_names() << ".bv\n";
	}
	
	std::cerr << "\nLog file is in\n";
	std::cerr << "\t" << log_path << "/" << A_file_manager->get_sum_of_file_names() << "_VS_" << B_file_manager->get_sum_of_file_names() << ".txt\n";
	std::cerr << "-------------------------------------------\n";
	std::cerr << "Total time: " << float(clock () - begin_time) / CLOCKS_PER_SEC << " s\n";
	
	std::cerr << "Similarity score between {";
	i = 0;
	for (; i < (int) B_file_names.size() - 1; i++) {
		std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << B_file_names.at(i).substr(B_file_names.at(i).rfind('/') + 1) << "} and {";
	i = 0;
	for (; i < (int) A_file_names.size() - 1; i++) {
		std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << ", ";
	}
	std::cerr << A_file_names.at(i).substr(A_file_names.at(i).rfind('/') + 1) << "}: " << (double) nb_common * 100.0 / (double) nb_total_reads << "%\n\n";
	return 0;
}

unsigned long compare (FileManager * index_file_manager, FileManager * search_file_manager, int kmer_size, int min_hits, unsigned long max_kmer)
{
	// -----------------------------------------------------------------------
	//                           INDEX & SEARCH
	// -----------------------------------------------------------------------
	BloomFilter bloom_filter (kmer_size);
	HashKey hash (kmer_size);
	std::string & current_read_to_index = index_file_manager->get_next_read();
	clock_t index_time = 0;
	clock_t search_time = 0;
	unsigned long nb_indexed_reads = 0;
	unsigned long nb_searched_reads = 0;
	unsigned long nb_found_reads = 0;
	while (!current_read_to_index.empty()) {
		// -----------------------------------------------------------------------
		//                                INDEX
		// -----------------------------------------------------------------------
		// Index reads from index_file_manager
		bloom_filter.clear();
		const clock_t begin_index = clock();
		unsigned long nb_indexed_kmers = 0;
		while (!current_read_to_index.empty() && nb_indexed_kmers < max_kmer) {
			nb_indexed_reads++;
			hash.clear();
			for (int i = 0; i < (int) current_read_to_index.size(); i++) {
				if (current_read_to_index[i] == 'N' || current_read_to_index[i] == 'n') {
					hash.clear();
				} else if (hash.add(current_read_to_index[i]) >= kmer_size) {
					bloom_filter.feed(hash);
					nb_indexed_kmers++;
				}
			}
			current_read_to_index = index_file_manager->get_next_read();
		}
		const clock_t end_index = clock();
		index_time += end_index - begin_index;
		if (nb_indexed_kmers == max_kmer) {
			std::cerr << "Max\n";
		}
		
		// -----------------------------------------------------------------------
		//                               SEARCH
		// -----------------------------------------------------------------------
		// Search reads from search_file_manager in the indexed reads
		nb_searched_reads = 0;
		search_file_manager->rewind();
		const clock_t begin_search = clock();
		std::string & current_read_to_search = search_file_manager->get_next_read();
		while (!current_read_to_search.empty()) {
			nb_searched_reads++;
			// Search k-mers
			int seen = 0;
			//search_file_manager->untag_current_read();
			bool found = false;
			hash.clear();
			for (long i = 0; i < (int) current_read_to_search.size() && !found; i++) {
				if (current_read_to_search[i] == 'N' || current_read_to_search[i] == 'n') {
					hash.clear();
				} else if (hash.add(current_read_to_search[i]) >= kmer_size) {
					if (bloom_filter.is_found(hash)) {
						seen++;
						if (seen >= min_hits) {
							search_file_manager->tag_current_read();
							nb_found_reads++;
							found = true;
						}
						hash.clear();
					}
				}
			}
			// Search the reverse strand if not found in the first step
			if (!found) {
				seen = 0;
				hash.clear();
				for (long i = 0; i < (int) current_read_to_search.size() && !found; i++) {
					if (current_read_to_search[i] == 'N' || current_read_to_search[i] == 'n') {
						hash.clear();
					} else if (hash.rv_add(current_read_to_search[i]) >= kmer_size) {
						if (bloom_filter.is_found(hash)) {
							seen++;
							if (seen >= min_hits) {
								search_file_manager->tag_current_read();
								nb_found_reads++;
								found = true;
							}
							hash.clear();
						}
					}
				}
			}
			current_read_to_search = search_file_manager->get_next_read();
		}
		const clock_t end_search = clock();
		search_time += end_search - begin_search;
	}
	
	std::cerr << "[" << nb_indexed_reads << " reads indexed, " << nb_searched_reads << " reads searched, " << nb_found_reads << " reads shared]\n";
	
	std::cerr << "Index time: " << float(index_time) / CLOCKS_PER_SEC << " s\n";
	std::cerr << "Query time: " << float(search_time) / CLOCKS_PER_SEC << " s\n";
	std::cerr << "Total time: " << float(index_time + search_time) / CLOCKS_PER_SEC << " s\n";
	
	return nb_found_reads;
}



// -----------------------------------------------------------------------
//                             PRINT USAGE
// -----------------------------------------------------------------------
void print_usage ()
{
	std::cerr << "\nCompare_reads, version " << version << "\n";
	std::cerr << "Usage : ./compare_reads -a <readsA.fasta>[,readsA.bv] -b <readsB.fastq>[,readsB.bv] [options]\n";
	std::cerr << "Mandatory:\n";
	std::cerr << "\t -a <readsA.fasta>[,readsA.bv]: First read set, in fasta or fastq, gzipped or not  - MANDATORY\n";
	std::cerr << "\t                                readsA.bv is the bit vector of the first read set  - OPTIONAL\n";
	std::cerr << "\t -b <readsB.fastq>[,readsB.bv]: Second read set, in fasta or fastq, gzipped or not - MANDATORY\n";
	std::cerr << "\t                                readsB.bv is the bit vector of the second read set - OPTIONAL\n";
	std::cerr << "Options:\n";
	std::cerr << "\t -l </.../>: ABSOLUTE path to log folder\n";
	std::cerr << "\t -o </.../>: ABSOLUTE path to output folder\n";
	std::cerr << "\t -k <value>: Size of k-mers (value of k) - DEFAULT = 32\n";
	std::cerr << "\t -t <value>: Minimum number of shared k-mers - DEFAULT = 2\n";
	std::cerr << "\t -m <value>: Maximum number of reads to use in each files - DEFAULT = all reads\n";
	std::cerr << "\t -h: Prints this message and exit\n";
	std::cerr << "\t -v: Prints the version number and exit\n";
}
